/* Copyright 2021, Christine Solnon, All rights reserved. */
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <getopt.h>
#include <string.h>

#define bool short
#define true 1
#define false 0

typedef struct{
    int nb;   // current number of values in the sparse set
    int* val; // val[0..nb-1] = current values in the sparse set
    int* pos; // pos[v] = position of v in val, i.e., val[i] = v <=> pos[v] = i
} SparseSet;

typedef struct{
    int n; // number of vertices
    int* nbAdj; // nbAdj[i] = number of vertices adjacent to i
    int** adj; // adj[i][j] = jth vertex adjacent to i
    char** isEdge; // isEdge[i][j] = 1 if (i,j) is an edge, and 0 otherwise
} Graph;

int iseed;

struct rusage ru;     // reusable structure to get CPU time usage

double randFloat(void) {
    /* Uniform random number generator x(n+1)= a*x(n) mod c
     with a = pow(7,5) and c = pow(2,31)-1.
     Copyright (c) Tao Pang 1997. */
    const int ia=16807,ic=2147483647,iq=127773,ir=2836;
    int il,ih,it;
    double rc;
    //extern int iseed;
    ih = iseed/iq;
    il = iseed%iq;
    it = ia*il-ir*ih;
    if (it > 0)    iseed = it;
    else iseed = ic+it;
    rc = ic;
    return(iseed/rc);
}

int getNextRand(int limit) {
    return (int)(100000*randFloat()) % limit;
}

int clique(int firstVertex, Graph* g, int* vertices){
    // Return the size of a clique built in a greedy way, starting from firstVertex
    // vertices[0..size-1] = vertices of the clique
    int size = 1;
    vertices[0] = firstVertex;
    int nbCand = g->nbAdj[firstVertex];
    int cand[nbCand];
    for (int i=0; i<nbCand; i++)
        cand[i] = g->adj[firstVertex][i];
    while (nbCand > 0){
        int i = getNextRand(nbCand);
        int v = cand[i];
        vertices[size++] = v;
        cand[i] = cand[--nbCand];
        int j=0;
        while (j<nbCand){
            if (g->isEdge[cand[j]][v]>0) j++;
            else cand[j] = cand[--nbCand];
        }
    }
    return size;
}

Graph* readGraph(char* fileName){
    // Read data in fileName and create the corresponding graph
    FILE* fd;
    if ((fd=fopen(fileName, "r"))==NULL){
        printf("ERROR: Cannot open ascii input file %s\n", fileName);
        exit(1);
    }
    Graph* g = (Graph*)malloc(sizeof(Graph));
    if (fscanf(fd,"%d",&(g->n)) != 1){
        printf("ERROR while reading input file %s\n", fileName);
        exit(1);
    }
    g->nbAdj = (int*)malloc(g->n*sizeof(int));
    g->adj = (int**)malloc(g->n*sizeof(int*));
    g->isEdge = (char**)malloc(g->n*sizeof(char*));
    for (int i=0; i<g->n; i++){
        g->isEdge[i] = (char*)calloc(g->n,sizeof(char));
        if (fscanf(fd,"%d",&(g->nbAdj[i])) != 1 || g->nbAdj[i] < 0 || feof(fd)){
            printf("ERROR while reading input file %s\n", fileName);
            exit(1);
        }
        g->adj[i] = (int*)malloc(g->nbAdj[i]*sizeof(int));
        for (int j=0; j<g->nbAdj[i]; j++){
            if (fscanf(fd,"%d",&(g->adj[i][j])) != 1 || g->adj[i][j] < 0 || g->adj[i][j] >= g->n || feof(fd)){
                printf("ERROR while reading input file %s\n", fileName);
                exit(1);
            }
            g->isEdge[i][g->adj[i][j]] = 1;
        }
    }
    return g;
}

void removeValue(int v, SparseSet* s){
    // remove v from s, i.e., exchange v with the last value in s
    s->nb--;
    s->val[s->pos[v]] = s->val[s->nb];
    s->pos[s->val[s->nb]] = s->pos[v];
    s->val[s->nb] = v;
    s->pos[v] = s->nb;
}

void removeAllValuesButOne(int v, SparseSet* s){
    // remove all values but v from s, i.e., exchange v with the first value in s
    s->val[s->pos[v]] = s->val[0];
    s->pos[s->val[0]] = s->pos[v];
    s->val[0] = v;
    s->pos[v] = 0;
    s->nb = 1;
}

bool isIn(int v, SparseSet* s){
    // return true is v is a value in s
    return (s->pos[v] < s->nb);
}

void initSparseSet(int size, SparseSet* s){
    // Initialize s with values 0..size-1
    s->val = (int*)malloc(size*sizeof(int));
    s->pos = (int*)malloc(size*sizeof(int));
    for (int i=0; i<size; i++){
        s->val[i] = i;
        s->pos[i] = i;
    }
    s->nb = size;
}

int nbNodes = 0; // Number of search nodes

bool propagate(int u, int c, SparseSet* uncolored, SparseSet* dom, int* deg, Graph* g){
    // propagate the fact that vertex u is colored with c
    nbNodes++;
    removeValue(u, uncolored);
    removeAllValuesButOne(c, &dom[u]);
    for (int i=0; i<g->nbAdj[u]; i++){
        int v = g->adj[u][i];
        if (isIn(v, uncolored)){
            deg[v]--;
            if (isIn(c,&dom[v])){
                if (dom[v].nb>1) removeValue(c, &dom[v]);
                else // no more possible colour for v
                    return false;
            }
            else if (deg[v] < dom[v].nb){
                // v can always be coloured
                removeValue(v, uncolored);
            }
        }
    }
    return true;
}

void printStats(bool feasible){
    getrusage(RUSAGE_SELF, &ru);
    if (feasible)
        printf("Feasible: ");
    else
        printf("Infeasible: ");
    printf("%d nodes; %d.%06d seconds\n", nbNodes, (int) ru.ru_utime.tv_sec, (int) ru.ru_utime.tv_usec);
}

void search(SparseSet* uncolored, SparseSet* dom, int* deg, Graph* g){
    // search for a proper colouring of g
    // uncoloured contains all vertices that are not yet coloured
    // for each vertex i, if i is coloured then dom[i] contains a single value which is the colour of i
    // otherwise, dom[i] contains all colours that may be assigned to i
    if (uncolored->nb == 0){ // solution found
        printStats(true);
        exit(0);
    }
    getrusage(RUSAGE_SELF, &ru);
    if ((int) ru.ru_utime.tv_sec > 1000){ // time limit reached
        printf("?: %d nodes; %d.%06d seconds\n", nbNodes, (int) ru.ru_utime.tv_sec, (int) ru.ru_utime.tv_usec);
        exit(0);
    }
    int next = -1;
    int minDom = g->n;
    int maxDeg = 0;
    int saveDomSize[g->n];
    int saveDeg[g->n];
    // search for the uncoloured vertex that has the smallest domain and break ties wrt degrees
    // (and save domain sizes and degrees)
    for (int i=0; i<uncolored->nb; i++){
        int u = uncolored->val[i];
        saveDomSize[u] = dom[u].nb;
        saveDeg[u] = deg[u];
        if (dom[u].nb < minDom || (dom[u].nb == minDom && g->nbAdj[u]>maxDeg)){
            next = u;
            minDom = dom[u].nb;
            maxDeg = g->nbAdj[u];
        }
    }
    // assign a colour to next
    if (minDom == 1){
        if (propagate(next, dom[next].val[0], uncolored, dom, deg, g))
            search(uncolored, dom, deg, g);
    }
    else{
        int saveDomNext[dom[next].nb];
        for (int i=0; i<dom[next].nb; i++) saveDomNext[i] = dom[next].val[i];
        int saveNbUncoloured = uncolored->nb;
        for (int i=0; i<saveDomSize[next]; i++){
            int c = saveDomNext[i];
            if (propagate(next, c, uncolored, dom, deg, g))
                search(uncolored, dom, deg, g);
            uncolored->nb = saveNbUncoloured;
            for (int j=0; j<uncolored->nb; j++){
                int u = uncolored->val[j];
                dom[u].nb = saveDomSize[u];
                deg[u] = saveDeg[u];
            };
        }
    }
}

void usage(int status){
    // print usage notice and exit with status code status
    printf("Usage:\n");
    printf("  -f FILE  Input graph (mandatory)\n");
    printf("  -t INT   Set CPU time limit in seconds (default: 60)\n");
    printf("  -k INT   Set the number of colours to k\n");
    printf("  -h       Print this help message\n");
    exit(status);
}

void parse(int* k, int* timeLimit, char* fileName, char* argv[], int argc){
    // get parameter values
    char ch;
    extern char* optarg;
    while ( (ch = getopt(argc, argv, "k:t:f:h"))!=-1 ) {
        switch(ch) {
            case 'f': strncpy(fileName, optarg, 254); break;
            case 't': *timeLimit=atoi(optarg); break;
            case 'k': *k=atoi(optarg); break;
            case 'h': usage(0);
            default:
                printf("Unknown option: %c.\n", ch);
                usage(2);
        }
    }
    if (fileName[0] == 0){
        printf("Error: no graph given.\n");
        usage(2);
    }
    if (k <= 0){
        printf("Error: k must be positive.\n");
        return usage(2);
    }
}

int main(int argc, char* argv[]){
    char fileName[1024];    // Name of the file that contains the graph
    int timeLimit=60;       // Default: CPU time limit set to 60 seconds
    int k;                  // Number of colours
    parse(&k, &timeLimit, fileName, argv, argc);
    Graph* g = readGraph(fileName);
    getrusage(RUSAGE_SELF, &ru);
    // For each vertex i, build a clique that contains i in a greedy way
    // Memorize the largest clique in maxClique, and its size in maxSize
    int maxSize = 0;
    int maxClique[g->n];
    int currentClique[g->n];
    for (int i=0; i<g->n; i++){
        int s = clique(i,g,currentClique);
        if (s>k){
            if (k<s){ // g cannot be coloured with less than s colours
                printStats(false);
                return 0;
            }
            maxSize = s;
            for (int j=0; j<s; j++)
                maxClique[j] = currentClique[j];
        }
    }
    // Initialize sparse sets for storing uncoloured vertices and available colours (dom)
    SparseSet uncolored;    // set of vertices that are not yet coloured
    initSparseSet(g->n, &uncolored);
    SparseSet dom[g->n];    // dom[u] = set of colours that may be assigned to u
    int deg[g->n];          // deg[u] = number of neighbours of u that are not yet coloured
    for (int i=0; i<g->n; i++){
        initSparseSet(k, &dom[i]);
        deg[i] = g->nbAdj[i];
        if (deg[i]<k) // i can always be coloured as it has less neighbours than its degree
            removeValue(i, &uncolored);
    }
    // break symmetries by colouring each vertex of the largest clique with a different colour
    for (int i=0; i<maxSize; i++){
        int u = maxClique[i];
        // color vertex u with color i
        if (isIn(u, &uncolored) && !propagate(u, i, &uncolored, dom, deg, g)){
            printStats(false);
            return 0;
        }
    }
    search(&uncolored, dom, deg, g);
    printStats(false);
    return 0;
}
